import numpy as np
from pint import Quantity

from rct229.utils.assertions import assert_


def aggregate_min_OA_schedule_across_zones(
    zone_OA_CFM_list_of_schedules: list[list[float]],
) -> list[Quantity]:
    """Each HVAC system can serve multiple zones which can have multiple terminal units serving each zone. In order to obtain an hourly HVAC level OA CFM schedule the OA CFM provided by each zone needs to be aggregated for each hour of the year. This function receives an RMD (B, U, or P) and a list of zone OA min cfm schedules (generated by the get_min_OA_CFM_sch_zone function) and loops through each zone to create an aggregated 8760 for OA CFM for the HVAC system.

    Parameters
    ----------
    zone_OA_CFM_list_of_schedules: list of list
        A list which contains each zone's 8760 OA CFM schedule as a list value.

    Returns
    -------
    list
        This function receives an RMD (B, U, or P) and a list of zone OA min cfm schedules (generated by the get_min_OA_CFM_sch_zone function) and loops through each zone to create an aggregated 8760 for OA CFM for the HVAC system.
    """
    assert_(
        all(
            [
                len(zone_OA_CFM_list_of_schedule) > 0
                for zone_OA_CFM_list_of_schedule in zone_OA_CFM_list_of_schedules
            ]
        ),
        "The provided list of schedules is an empty list.",
    )

    schedule_len = len(next(iter(zone_OA_CFM_list_of_schedules)))
    assert_(
        all(
            len(schedule) == schedule_len and len(schedule) in (8760, 8784)
            for schedule in zone_OA_CFM_list_of_schedules
        ),
        "Not all schedules length is equal to 8760 or 8784 or equal each other.",
    )

    if len(zone_OA_CFM_list_of_schedules) == 1:
        # one schedule sceanrio
        hourly_zone_running_OA_CFM_total = zone_OA_CFM_list_of_schedules[0]
    else:
        # multi-schedule scenario
        magnitudes = [
            [quantity.magnitude for quantity in sublist]
            for sublist in zone_OA_CFM_list_of_schedules
        ]
        units = zone_OA_CFM_list_of_schedules[0][0].units

        zone_oa_cfm_magnitudes_array = np.array(magnitudes)
        zone_oa_cfm_magnitudes_sum = np.sum(zone_oa_cfm_magnitudes_array, axis=0)

        hourly_zone_running_OA_CFM_total = [
            magnitude * units for magnitude in zone_oa_cfm_magnitudes_sum
        ]

    return hourly_zone_running_OA_CFM_total
